#ifndef LIBRAPID_TEST
#define LIBRAPID_TEST

#include <librapid/config.hpp>

namespace librapid::test {
	int testLibrapid(int x);

	void streamTest();
} // namespace librapid::test

#endif // LIBRAPID_TEST